#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2018 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""
This implements :class:`gpxity.WPTrackServer`.

WPTrackserver talks directly to the WP mysql database holding the trackserver data.

This backend does not directly support Track.category, Track.status, Track.keywords.
All this is added to the description when writing and extracted when reading. So
the description in the backend will contain all keywords, but Track.description
does not.

The database has a maximum field length of 255 for strings. If it is exceeded, the
last part of the strings will be thrown away silently when writing into the database.

"""

# pylint: disable=protected-access

import datetime

from gpxpy import gpx as mod_gpx

from ..backend import Backend
from .mmt import MMT

try:
    import MySQLdb
    HAVE_MYSQL = True
    import _mysql_exceptions
except ImportError:
    HAVE_MYSQL = False

__all__ = ['WPTrackserver']


class WPTrackserver(Backend):

    """Talk directly to the wordpress mysql database holding the trackserver data.

    See https://www.grendelman.net/wp/trackserver-wordpress-plugin/

    The section in  auth.cfg could look like

    [wptrackserver:username]
    Url = hostname
    Mysql =user@db_name
    Password = mysql_password

    username is the wordpress user name.

    hostname is the server wordpress is running on

    user is the mysql user. Find it in the wordpress config file,
    look for DB_USER.

    db_name is the name of the database. Find it in the wordpress config file,
    look for DB_NAME.

    mysql_password is for the mysql user. Find it in the wordpress config file,
    look for DB_PASSWORD.

    """

    # pylint: disable=abstract-method

    legal_categories = MMT.legal_categories

    needs_config = False

    _keywords_marker = '\nKEYWORDS: '

    _max_length = {'title': 255, 'description': 255}

    def __init__(self, url=None, auth=None, cleanup=False, timeout=None):
        """See class docstring. The url is host."""
        super(WPTrackserver, self).__init__(url, auth, cleanup, timeout)
        self.logger.debug('new WPTrackserver:%s', id(self))
        try:
            user, database = self.config.mysql.split('@')
        except ValueError:
            raise Backend.BackendException('Url is illegal: {}'.format(self.url))
        self.logger.debug('connecting to %s as %s with pw %s to db %s', self.url, user, self.config.password, database)
        try:
            self._db = MySQLdb.connect(
                host=self.url, user=user, passwd=self.config.password, database=database,
                autocommit=True, charset='utf8')
        except _mysql_exceptions.OperationalError as exc:
            raise Backend.BackendException(exc)
        self._cursor = self._db.cursor()
        self._cursor.execute('select id from wp_users where user_login=%s', [self.config.username])
        row = self._cursor.fetchone()
        if row is None:
            raise Backend.BackendException('WPTrackserver: User {} is not known'.format(self.config.username))
        self.user_id = row[0]

    def _encode_description(self, track):
        """Encode keywords in description.

        If description exceeds its maximum length, first remove keywords and then
        shorten the description.

        Returns: The string to be saved in the database.

        """
        def fmt_result():
            return '{}{}{}'.format(
                track.description, self._keywords_marker, ', '.join(kw_parts))

        max_length = self._max_length['description']
        kw_parts = track._encode_keywords().split(', ')
        result = fmt_result()
        if len(result) > max_length:
            while kw_parts and len(result) > max_length:
                kw_parts = kw_parts[:-1]
                result = fmt_result()
        return result[:max_length]

    def _decode_description(self, track, value, into_header_data=False):
        """Extract keywords.

        Returns: The decoded description

        """
        if self._keywords_marker in value:
            descr, raw_keywords = value.split(self._keywords_marker)
            # not into_header_data! because _read_all only delivers points
            track._decode_keywords(raw_keywords)
        else:
            descr = value
        track._header_data['description'] = descr
        return descr

    def _enrich_with_headers(self, track, row):
        """Get header values out of row."""
        track._header_data['time'] = row[1]
        track._header_data['title'] = row[2]
        self._decode_description(track, row[3])
        track._header_data['distance'] = row[4] / 1000.0

    def _yield_tracks(self):
        """."""
        self._cursor.execute(
            'select id,created,name,comment,distance from wp_ts_tracks where user_id=%s', [self.user_id])
        for _ in self._cursor.fetchall():
            track = self._found_track('{:06}'.format(_[0]))
            self._enrich_with_headers(track, _)
            yield track

    @staticmethod
    def __point(row):
        """Make a GPX point.

        Returns:
            The point

        """
        return mod_gpx.GPXTrackPoint(
            latitude=float(row[0]),
            longitude=float(row[1]),
            time=row[2])

    def _read_all(self, track) ->None:
        """Read the full track."""
        assert track.id_in_backend
        self._cursor.execute(
            'select latitude,longitude,occurred from wp_ts_locations where trip_id=%s', [track.id_in_backend])
        track.add_points([self.__point(x) for x in self._cursor.fetchall()])

    def _save_header(self, track):
        """Write all header fields. May set track.id_in_backend."""
        description = self._encode_description(track)
        title = track.title[:self._max_length['title']]
        track_time = track.time or datetime.datetime(year=1970, month=1, day=1, hour=1)
        if track.id_in_backend is None:
            self._cursor.execute(
                'insert into wp_ts_tracks(user_id,name,created,comment,distance,source) values(%s,%s,%s,%s,%s,%s)',
                (self.user_id, title, track_time, description, track.distance(), ''))
            track.id_in_backend = str(self._cursor.lastrowid)
            self.logger.debug(
                'new id %s: insert into wp_ts_tracks(user_id,name,created,comment,distance,source) '
                'values(%s,%s,%s,%s,%s,%s)',
                track.id_in_backend, self.user_id, title, track_time, description, track.distance(), '')
        else:
            self._cursor.execute(
                'update wp_ts_tracks set name=%s,created=%s,comment=%s,distance=%s where id=%s',
                (title, track_time, description, track.distance(), track.id_in_backend))
            self.logger.debug(
                'update wp_ts_tracks set name=%s,created=%s,comment=%s,distance=%s where id=%s',
                title, track_time, description, track.distance(), track.id_in_backend)
            self.logger.debug('save_header: rewrite %s', track.id_in_backend)

    def _write_all(self, track) ->str:
        """save full gpx track.

        Since the file name uses title and title may have changed,
        compute new file name and remove the old files. We also adapt track.id_in_backend.

        Returns:
            the new track.id_in_backend

        """
        self._save_header(track)
        self._cursor.execute('delete from wp_ts_locations where trip_id=%s', [track.id_in_backend])
        data = [(track.id_in_backend, x.latitude, x.longitude, x.elevation or 0.0, x.time)
                for x in track.points()]  # noqa
        self._cursor.executemany(
            'insert into wp_ts_locations(trip_id, latitude, longitude, altitude, occurred, comment, speed, heading)'
            ' values(%s, %s, %s, %s, %s,"",0.0, 0.0)', data)
        return track.id_in_backend

    def _remove_ident(self, ident: str) ->None:
        """backend dependent implementation."""
        self._cursor.execute('delete from wp_ts_locations where trip_id=%s', [ident])
        cmd = 'delete from wp_ts_tracks where id=%s'
        self.logger.debug(cmd, ident)
        self._cursor.execute(cmd, [ident])

    @classmethod
    def is_disabled(cls) ->bool:
        """True if this backend is disabled by env variable GPXITY_DISABLE_BACKENDS.

        or because python mysql is not installed.

        This variable is a comma separated list of Backend class names.

        Returns:
            True if disabled

        """
        if not HAVE_MYSQL:
            return True
        return super(WPTrackserver, cls).is_disabled()



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""This module defines :class:`~gpxity.Authenticate`."""

import os
import logging
from configparser import ConfigParser

__all__ = ['Authenticate']


class Authenticate:

    """
    Get password and / or Url from auth.cfg.

    If nothing is useable, sets them to None.

    auth.cfg is expected in :literal:`~/.config/Gpxity/auth.cfg`

    .. DANGER::
       auth.cfg is not encrypted. Better not use this unless you know what you are doing!

    Args:
        backend (Backend): The backend
        username (str): For the wanted account in the backend

    Attributes:
        path (str): The name for the auth file. Class variable, to be changed before
          Authenticate() is instantiated.

        auth (tuple(str,str)): (username, password). Both are either str or None.

        url: If given, overrides the url given to the backend

    For every specific account in a backend, auth.cfg has a section. Its name is
    case sensitive: The ClassName must match exactly.

      * [ClassName.username]

    A section can define
      * Password
      * Url
      * Mysql, used by WPTrackserver

    An example for user gpxitytest on gpsies.com:

    .. code-block:: guess

        [GPSIES:gpxitytest]
        Password = the_unencrypted_password

    A mail account:

    .. code-block:: guess

        [Mailer:gpxitytest]
        Url = tester@test.test


    """

    # pylint: disable=too-few-public-methods

    path = '~/.config/Gpxity/auth.cfg'

    def __init__(self, backend, url, username: str = None):
        """See class docstring."""

        self.__config = ConfigParser()
        if isinstance(username, dict):
            self.__section_name = '{}:from_dict'.format(backend.__class__.__name__)
            self.__config[self.__section_name] = {}
            self.section = self.__config[self.__section_name]
            for key, value in username.items():
                self.section[key] = str(value)
        else:
            self.__section_name = '{}:{}'.format(backend.__class__.__name__, username)
            self.__path = os.path.expanduser(self.path)
            if os.path.exists(self.__path):
                with open(self.__path) as auth_file:
                    self.__config.read_string(auth_file.read())
                    if self.__section_name not in self.__config:
                        if backend.needs_config:
                            raise KeyError('Section [{}] not found in {}'.format(self.__section_name, self.__path))
                        self.__config[self.__section_name] = {}
                    self.section = self.__config[self.__section_name]
                    self.section['username'] = username or ''
            else:
                logging.info('%s not found', self.__path)
                self.section = dict()
        if self.section.get('username', None) == 'wrong_user':
            raise KeyError
        if 'url' not in self.section:
            self.section['url'] = url or ''
        self._check_url()

    def _check_url(self):
        """Check syntax for url."""
        _ = self.url
        if _ and _.endswith('/') and _ != '/':
            raise Exception('url {} must not end with /'.format(_))

    def __getattr__(self, key):
        """Only called if key is not an existing attribute.

        Returns: The value

        """
        return self.section[key] if key in self.section else None



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""implements :class:`gpxity.backends.test.test_backends.TestBackends` for all backends."""

# pylint: disable=protected-access

import os
import time
import datetime
import random
import tempfile

from unittest import skipIf

from .basic import BasicTest, disabled
from .. import Directory, MMT, GPSIES, ServerDirectory, TrackMMT, Mailer, WPTrackserver
from ... import Track, Lifetrack, Backend

# pylint: disable=attribute-defined-outside-init


class TestBackends(BasicTest):

    """Are the :literal:`supported_` attributes set correctly?."""

    def test_supported(self):
        """Check values in supported for all backends."""
        expect_unsupported = dict()
        expect_unsupported[Directory] = {
            'lifetrack', 'lifetrack_end', 'write_add_keywords', 'write_remove_keywords',
            'write_category', 'write_description', 'write_public', 'write_title'}
        expect_unsupported[ServerDirectory] = {
            'lifetrack', 'lifetrack_end', 'write_add_keywords', 'write_remove_keywords',
            'write_category', 'write_title', 'write_description', 'write_public'}
        expect_unsupported[MMT] = set()
        expect_unsupported[GPSIES] = {
            'get_time', 'lifetrack', 'lifetrack_end', 'keywords', 'write_add_keywords', 'write_remove_keywords'}
        expect_unsupported[TrackMMT] = {
            'scan', 'remove',
            'write_title', 'write_description', 'write_public',
            'write_category', 'write_add_keywords',
            'write_remove_keywords'}
        expect_unsupported[Mailer] = {
            'scan', 'remove', 'get_time',
            'write_title', 'write_description', 'write_public',
            'write_category', 'write_add_keywords',
            'write_remove_keywords', 'lifetrack'}
        expect_unsupported[WPTrackserver] = {
            'get_time', 'lifetrack', 'lifetrack_end',
            'write_add_keywords', 'write_remove_keywords', 'write_category',
            'write_description', 'write_public', 'write_title'}
        for cls in Backend.all_backend_classes():
            with self.subTest(cls):
                self.assertTrue(cls.supported & expect_unsupported[cls] == set())
                self.assertEqual(sorted(cls.supported | expect_unsupported[cls]), sorted(cls.full_support))

    def test_all_backends(self):
        """Check if Backend.all_backend_classes works."""
        backends = Backend.all_backend_classes()
        expected = [Directory, GPSIES, MMT, Mailer, ServerDirectory, TrackMMT, WPTrackserver]
        expected = [x for x in expected if not x.is_disabled()]
        self.assertEqual(backends, expected)

    def test_save_empty(self):
        """Save empty track."""
        for cls in Backend.all_backend_classes(needs={'write'}):
            with self.subTest(cls):
                can_remove = 'remove' in cls.supported
                with self.temp_backend(cls, cleanup=can_remove, clear_first=can_remove) as backend:
                    track = Track()
                    if cls in (MMT, TrackMMT, GPSIES):
                        with self.assertRaises(cls.BackendException):
                            backend.add(track)
                    else:
                        self.assertIsNotNone(backend.add(track))

    @skipIf(*disabled(Directory))
    def test_directory_backend(self):
        """Manipulate backend."""
        track = self.create_test_track()
        with Directory(cleanup=True) as directory1:
            with Directory(cleanup=True) as directory2:
                saved = directory1.add(track)
                self.assertEqual(len(directory1), 1)
                self.assertEqual(saved.backend, directory1)
                directory1.add(track.clone())
                self.assertEqual(len(directory1), 2)
                directory2.add(track)
                self.assertEqual(len(directory2), 1)
                directory2.scan()
                self.assertEqual(len(directory2), 1)

    def test_slow_duplicate_tracks(self):
        """What happens if we save the same track twice?."""
        for cls in Backend.all_backend_classes(needs={'remove', 'write'}):
            with self.subTest(cls):
                with self.temp_backend(cls) as backend:
                    track = self.create_test_track()
                    backend.add(track)
                    self.assertEqual(len(backend), 1)
                    with self.assertRaises(ValueError):
                        backend.add(track)
                    self.assertEqual(len(backend), 1)
                    if cls is GPSIES:
                        # if the same track data is uploaded again, we get the same id_in_backend.
                        with self.assertRaises(ValueError):
                            backend.add(track.clone())
                        self.assertEqual(len(backend), 1)
                    else:
                        backend.add(track.clone())
                        self.assertEqual(len(backend), 2)

    def test_open_wrong_username(self):
        """Open backends with username missing in auth.cfg."""
        for cls in Backend.all_backend_classes(exclude=[Directory, ServerDirectory]):
            with self.subTest(cls):
                with self.assertRaises(KeyError):
                    self.setup_backend(cls, username='wrong_user')

    def test_open_wrong_password(self):
        """Open backends with wrong password."""
        for cls in Backend.all_backend_classes(needs={'scan'}):
            with self.subTest(cls):
                if not issubclass(cls, Directory):
                    with self.assertRaises(cls.BackendException):
                        self.setup_backend(cls, username='wrong_password')

    @skipIf(*disabled(Directory))
    def test_match(self):
        """test backend match function.

        Returns:
            None

        """
        def match_date(track) ->str:
            """match against a date.

            Returns:
                None if match else an error message
            """
            if track.time < datetime.datetime(year=2016, month=9, day=5):
                return 'time {} is before {}'.format(track.time, '2016-09-05')
            return None
        cls = Directory
        with self.subTest(cls):
            with self.temp_backend(cls, count=3) as backend:
                for idx, _ in enumerate(backend):
                    _.adjust_time(datetime.timedelta(hours=idx))
                new_track = backend[0].clone()
                self.assertIsNotNone(match_date(new_track))
                self.assertEqual(len(backend), 3)
                backend.match = match_date
                self.assertEqual(len(backend), 1)
                with self.assertRaises(cls.NoMatch):
                    backend.add(new_track)
                self.assertEqual(len(backend), 1)
                orig_time = backend[0].time
                delta = datetime.timedelta(days=-5)
                with self.assertRaises(cls.NoMatch):
                    backend[0].adjust_time(delta)
                self.assertEqual(len(backend), 1)
                self.assertEqual(orig_time + delta, backend[0].time)

    def test_z9_create_backend(self):
        """Test creation of a backend."""
        for cls in Backend.all_backend_classes(needs={'remove', 'get_time'}):
            with self.subTest(cls):
                with self.temp_backend(cls, count=3) as backend:
                    self.assertEqual(len(backend), 3)
                    first_time = backend.get_time()
                    time.sleep(2)
                    second_time = backend.get_time()
                    total_seconds = (second_time - first_time).total_seconds()
                    self.assertTrue(1 < total_seconds < 8, 'Time difference should be {}, is {}-{}={}'.format(
                        2, second_time, first_time, second_time - first_time))

    def test_slow_write_remoteattr(self):
        """If we change title, description, public, category in track, is the backend updated?."""
        for cls in Backend.all_backend_classes(needs={'remove', }):
            with self.subTest(cls):
                with self.temp_backend(cls, count=1, category='Horse riding') as backend:
                    track = backend[0]
                    first_public = track.public
                    first_title = track.title
                    first_description = track.description
                    first_category = track.category
                    self.assertEqual(first_category, 'Horse riding')
                    self.assertFalse(track.public)
                    track.public = True
                    self.assertTrue(track.public)
                    track.title = 'A new title'
                    self.assertEqual(track.title, 'A new title')
                    track.description = 'A new description'
                    track.category = 'Cycling'
                    # make sure there is no cache in the way
                    backend2 = backend.clone()
                    track2 = backend2[0]
                    self.assertEqualTracks(track, track2, with_category=False)
                    self.assertNotEqual(first_public, track2.public)
                    self.assertNotEqual(first_title, track2.title)
                    self.assertNotEqual(first_description, track2.description)
                    self.assertNotEqual(first_category, track2.category)

    def xtest_gpsies_bug(self):
        """We have this bug only sometimes: title, category or time will be wrong in track2.
        Workaround is in GPSIES._edit."""
        for _ in range(20):
            with self.temp_backend(GPSIES, count=1, category='Horse riding') as backend:
                track = backend[0]
                track.title = 'A new title'
                track.description = 'A new description'
                track.category = 'Cycling'
                # make sure there is no cache in the way
                backend2 = backend.clone()
                track2 = backend2[0]
                self.assertEqualTracks(track, track2, with_category=True)

    def test_z2_keywords(self):
        """save and load keywords.

        For now, all test keywords start with uppercase, avoiding MMT problems
        """         # noqa hides bug in eric6 style checker

        kw_a = 'A'
        kw_b = 'Berlin'
        kw_c = 'CamelCase'
        kw_d = self.unicode_string2

        def minus(value):
            """

            Returns:
                value preceded with -

            """
            return '-' + value

        for cls in Backend.all_backend_classes(needs={'write', 'scan', 'keywords'}):
            with self.subTest(cls):
                is_mmt = cls.__name__ == 'MMT'
                with self.temp_backend(cls, clear_first=not is_mmt, cleanup=not is_mmt) as backend:
                    if not backend:
                        continue
                    track = backend[0]
                    track.keywords = list()
                    self.assertEqual(track.keywords, list())
                    track.keywords = ([kw_a, kw_b, kw_c])
                    track.change_keywords(minus(kw_b))
                    self.assertEqual(track.keywords, ([kw_a, kw_c]))
                    with self.assertRaises(Exception):
                        track.change_keywords('Category:whatever')
                    track.change_keywords(kw_d)
                    self.assertEqual(set(track.keywords), {kw_a, kw_c, kw_d})
                    backend2 = backend.clone()
                    track2 = backend2[track.id_in_backend]
                    track2.change_keywords(minus(kw_d))
                    self.assertEqual(track2.keywords, ([kw_a, kw_c]))
                    self.assertEqual(track.keywords, ([kw_a, kw_c, kw_d]))
                    backend.scan()
                    self.assertEqual(track.keywords, ([kw_a, kw_c, kw_d]))
                    self.assertEqual(backend[track.id_in_backend].keywords, ([kw_a, kw_c]))
                    track.change_keywords(minus(kw_a))
                    # this is tricky: The current implementation assumes that track.keywords is
                    # current - which it is not. track still thinks kw_d is there but it has been
                    # removed by somebody else. MMT has a work-around for removing tracks which
                    # removes them all and re-adds all wanted. So we get kw_d back.
                    self.assertEqual(track.keywords, ([kw_c, kw_d]))
                    # track2.change_keywords(minus(kw_a))
                    track.change_keywords(minus(kw_c))
                    track.change_keywords(minus(kw_d))
                    backend.scan()
                    self.assertEqual(backend[0].keywords, list())

    def test_z_unicode(self):
        """Can we up- and download unicode characters in all text attributes?."""
        tstdescr = 'DESCRIPTION with ' + self.unicode_string1 + ' and ' + self.unicode_string2
        for cls in Backend.all_backend_classes(needs={'remove'}):
            with self.subTest(cls):
                with self.temp_backend(cls, count=1) as backend:
                    backend2 = backend.clone()
                    track = backend[0]
                    self.assertIsNotNone(track.backend)
                    track.title = 'Title ' + self.unicode_string1
                    self.assertIsNotNone(track.backend)
                    self.assertEqual(track.backend, backend)
                    backend2.scan()  # because backend2 does not know about changes thru backend
                    track2 = backend2[0]
                    # track and track2 may not be identical. If the original track
                    # contains gpx xml data ignored by MMT, it will not be in track2.
                    self.assertEqual(track.title, track2.title)
                    track.description = tstdescr
                    self.assertEqual(track.description, tstdescr)
                    if cls is Directory:
                        self.assertTrackFileContains(track, tstdescr)
                    backend2.scan()
                    self.assertEqual(backend2[0].description, tstdescr)
                    backend2.destroy()

    def test_change_points(self):
        """Can we change the points of a track?.

        For MMT this means re-uploading and removing the previous instance, so this

        is not always as trivial as it should be."""

    @skipIf(*disabled(MMT))
    def test_slow_download_many(self):
        """Download many tracks."""
        many = 150
        backend = self.setup_backend(MMT, username='gpxstoragemany', count=many, cleanup=False, clear_first=True)
        self.assertEqual(len(backend), many)

    def test_duplicate_title(self):
        """two tracks having the same title."""
        for cls in Backend.all_backend_classes(needs={'remove'}):
            with self.subTest(cls):
                with self.temp_backend(cls, count=2) as backend:
                    backend[0].title = 'TITLE'
                    backend[1].title = 'TITLE'

    @skipIf(*disabled(Directory))
    def test_private(self):
        """Up- and download private tracks."""
        with self.temp_backend(Directory, count=5, category='Cycling') as local:
            track = Track(gpx=self._get_gpx_from_test_file('test2'))
            self.assertTrue(track.public)  # as defined in test2.gpx keywords
            track.public = False
            self.assertFalse(track.public)
            local.add(track)
            for cls in Backend.all_backend_classes(needs={'remove'}):
                with self.subTest(cls):
                    with self.temp_backend(cls) as backend:
                        backend.merge(local)
                        for _ in backend:
                            self.assertFalse(_.public)
                        backend2 = backend.clone()
                        with Directory(cleanup=True) as copy:
                            for _ in copy.merge(backend2):
                                self.logger.debug(_)
                            self.assertSameTracks(local, copy)

    @skipIf(*disabled(Directory))
    def test_merge_backends(self):
        """merge backends."""
        def dump(msg):
            for line in msg:
                self.logger.debug(line)
        org_source_len = 2
        assert org_source_len >= 2
        org_sink_len = 2
        assert org_sink_len >= 2
        with self.temp_backend(Directory, url='source', count=org_source_len) as source:
            with self.temp_backend(Directory, url='sink', count=org_sink_len) as sink:
                for _ in list(sink)[1:]:
                    _.adjust_time(datetime.timedelta(hours=100))

                # let's have two identical tracks in source without match in sink:
                next(source[0].points()).latitude = 6
                source.add(source[0].clone())

                # and two identical tracks in sink without match in source:
                next(sink[1].points()).latitude = 7
                sink.add(sink[1].clone())

                # and one track twice in source and once in sink:
                sink.add(source[0])

                # and one track once in source and once in sink:
                sink.add(source[1])

                self.assertEqual(len(source), org_source_len + 1)
                self.assertEqual(len(sink), org_sink_len + 3)

                dump(sink.merge(source, dry_run=True))
                self.assertEqual(len(source), org_source_len + 1)
                self.assertEqual(len(sink), org_sink_len + 3)

                dump(sink.merge(source))
                self.assertEqual(len(source), org_source_len + 1)
                self.assertEqual(len(sink), org_source_len + org_sink_len + 1)

                for _ in range(2):
                    dump(sink.merge(source, remove=True))
                    self.assertEqual(len(source), 0)
                    self.assertEqual(len(sink), org_source_len + org_sink_len)

    @skipIf(*disabled(Directory))
    def test_scan(self):
        """some tests about Backend.scan()."""
        with self.temp_backend(Directory, count=5) as source:
            backend2 = source.clone()
            track = self.create_test_track()
            backend2.add(track)
            self.assertEqual(len(backend2), 6)
            source.scan()  # because it cannot know backend2 added something

    @skipIf(*disabled(MMT))
    def test_lifetrack_mmt(self):
        """test life tracking against a free account on mapmytracks.com."""
        with MMT(auth='gpxitytest') as uplink:
            self.assertTrue(uplink.is_free_account)
            life = Lifetrack(uplink)
            with self.assertRaises(Exception) as context:
                life.start(self._random_points())
            self.assertEqual(str(context.exception), 'Your free MMT account does not allow lifetracking')

    @skipIf(*disabled(Directory, TrackMMT))
    def test_lifetrack_local(self):
        """test life tracking against a local server."""
        def track(*args):
            life = Lifetrack(*args)
            points = self._random_points(100)
            life.start(points[:50])
            time.sleep(7)
            life.update(points[50:])
            life.end()
            for target in life.targets:
                track = target.track
                new_id = track.id_in_backend
                if 'scan' in target.backend.supported:
                    self.assertIn(new_id, uplink)
                    self.assertSameTracks(uplink, serverdirectory)
                else:
                    with self.assertRaises(NotImplementedError):
                        new_id in target.backend  # pylint: disable=pointless-statement

        with self.temp_backend(Directory) as serverdirectory:
            with self.lifetrackserver(servername='localhost', port=12398, directory=serverdirectory.url):
                with TrackMMT(auth='gpxitytest') as uplink:
                    if Mailer.is_disabled():
                        track(uplink)
                    else:
                        with self.temp_backend(Mailer) as mailer:
                            mailer.min_interval = 5
                            track(uplink, mailer)
                            self.assertEqual(len(mailer.history), 3)
                            self.assertIn('Lifetracking starts', mailer.history[0])
                            self.assertIn('Lifetracking continues', mailer.history[1])
                            self.assertIn('Lifetracking ends', mailer.history[2])

    def test_backend_dirty(self):
        """Track._dirty."""
        for cls in Backend.all_backend_classes(needs={'scan', 'write'}):
            with self.subTest(cls):
                with self.temp_backend(cls, count=1) as backend:
                    track = backend[0]
                    with self.assertRaises(Exception):
                        track._dirty = False
                    self.assertFalse(track._dirty)
                    # version 1.1 should perhaps be a test on its own, see Track.to_xml()
                    track.category = 'Driving'
                    track._dirty = 'gpx'
                    self.assertFalse(track._dirty)
                    backend2 = backend.clone()
                    self.assertEqual(backend2[0].category, 'Driving')
                    b2track = backend2[0]
                    self.assertEqual(b2track.category, 'Driving')
                    b2track.title = 'another new title'
                    self.assertEqual(b2track.category, 'Driving')
                    self.assertEqual(backend2[0].category, 'Driving')
                    track.title = 'new title'
                    self.assertEqual(track.category, 'Driving')
                    backend3 = backend.clone()
                    self.assertEqual(backend3[0].category, 'Driving')
                    self.assertFalse(track._dirty)
                    with track.batch_changes():
                        track.title = 'new 2'
                        self.assertEqual(track._dirty, ['title'])
                    self.assertFalse(track._dirty)
                    with track.batch_changes():
                        track.title = 'new 3'
                        track.keywords = ['Something', 'something xlse']
                    backend4 = backend.clone()
                    self.assertEqual(backend4[0].title, 'new 3')

    def test_directory_dirty(self):
        """test gpx._dirty where id_in_backend is not the default.

        Currently track._dirty = 'gpx' changes the file name which is wrong."""

    @skipIf(*disabled(Directory))
    def test_directory(self):
        """directory creation/deletion."""

        dir_a = Directory(cleanup=True)
        self.assertTrue(dir_a.is_temporary)
        a_url = dir_a.url
        self.assertTrue(os.path.exists(a_url))
        dir_a.destroy()
        self.assertFalse(os.path.exists(a_url))

        test_url = tempfile.mkdtemp()
        dir_b = Directory(url=test_url, cleanup=True)
        self.assertFalse(dir_b.is_temporary)
        self.assertTrue(dir_b.url == test_url)
        dir_b.destroy()
        self.assertTrue(os.path.exists(test_url))
        os.rmdir(test_url)

        dir_c = Directory(auth='gpxitytest')
        self.assertTrue(dir_c.is_temporary)

        self.assertIn('/gpxity.TestBackends.test_directory_', dir_c.url)
        dir_c.destroy()

    @skipIf(*disabled(MMT))
    def test_mmt_empty(self):
        """MMT refuses upload without a specific error message if there is no track point."""
        track = self.create_test_track()
        del track.gpx.tracks[0]
        with MMT(auth='gpxitytest', cleanup=True) as mmt:
            with self.assertRaises(mmt.BackendException):
                mmt.add(track)

    def test_setters(self):
        """For all Track attributes with setters, test if we can change them without changing something else."""
        for cls in Backend.all_backend_classes(needs={'write', 'scan'}):
            with self.subTest(cls):
                with self.temp_backend(cls, count=1) as backend:
                    track = backend[0]
                    backend2 = backend.clone()
                    self.assertEqualTracks(track, backend2[0], with_category=False)
                    test_values = {
                        'title': ('default title', 'Täst Titel'),
                        'description': ('default description', 'Täst description'),
                        'category': ('Driving', 'Rowing'), 'public': (True, False)}
                    if cls is not GPSIES:
                        test_values['keywords'] = (['A', 'Hello Dolly', 'Whatever'], ['Something Else', 'Two'])
                    for main in test_values:
                        for key, (default_value, _) in test_values.items():
                            if key != main:
                                setattr(track, key, default_value)
                        setattr(track, main, test_values[main][1])
                        backend2.scan()
                        for key, (default_value, _) in test_values.items():
                            if key != main:
                                self.assertEqual(getattr(backend2[0], key), default_value)
                        self.assertEqual(getattr(backend2[0], main), test_values[main][1])

    def test_keywords(self) ->None:
        """Test arbitrary keyword changes.

        Returns:
            None

        """
        for cls in Backend.all_backend_classes(needs={'scan', 'keywords'}):
            with self.subTest(cls):
                with self.temp_backend(cls, count=1) as backend:
                    backend2 = backend.clone()
                    track = backend[0]
                    keywords = {
                        backend._encode_keyword(x)
                        for x in self._random_keywords(count=50)}
                    for _ in range(20):
                        self.assertEqual(backend._get_current_keywords(track), track.keywords)
                        add_keywords = set(random.sample(keywords, random.randint(0, 10)))
                        remove_keywords = set(random.sample(keywords, random.randint(0, 10)))
                        if not add_keywords & remove_keywords:
                            continue
                        expected_keywords = (set(track.keywords) | add_keywords) - remove_keywords
                        track.change_keywords(list(add_keywords) * 2)
                        self.assertEqual(
                            backend._get_current_keywords(track),
                            sorted(list(set(track.keywords) | add_keywords)))
                        track.change_keywords('-' + x for x in remove_keywords)
                        self.assertEqual(backend._get_current_keywords(track), sorted(expected_keywords))
                        self.assertEqual(sorted(expected_keywords), sorted(track.keywords))
                        backend2.scan()
                        self.assertEqual(sorted(expected_keywords), backend2[0].keywords)
                    with track.batch_changes():
                        # WPTrackserver has limited field lengths
                        loops, kwcount = (5, 5) if cls is WPTrackserver else (50, 10)
                        for _ in range(loops):
                            add_keywords = set(random.sample(keywords, random.randint(0, kwcount)))
                            remove_keywords = set(random.sample(keywords, random.randint(0, kwcount))) & add_keywords
                            if not add_keywords & remove_keywords:
                                continue
                            expected_keywords = (set(track.keywords) | add_keywords) - remove_keywords
                            track.change_keywords(add_keywords)
                            track.change_keywords('-' + x for x in remove_keywords)
                            self.assertEqual(sorted(expected_keywords), sorted(track.keywords))
                    backend2.scan()
                    self.assertEqual(
                        backend2._get_current_keywords(backend2[0]),
                        backend2[0].keywords)
                    self.assertEqual(sorted(expected_keywords), backend2[0].keywords)

    @skipIf(*disabled(Directory))
    def test_legal_categories(self):
        """Check if our fixed list of categories still matches the online service."""
        for cls in (MMT, GPSIES, TrackMMT):
            if cls.is_disabled:
                continue
            with self.subTest(cls), self.temp_backend(Directory) as serverdirectory:
                with self.temp_backend(cls, clear_first=False, cleanup=False) as backend:
                    if cls is TrackMMT:
                        if not Mailer.is_disabled():
                            # gpxity_server needs Mailer
                            with self.lifetrackserver(
                                    servername='localhost', port=12398, directory=serverdirectory.url):
                                downloaded = backend._download_legal_categories()
                                self.assertEqual(sorted(backend.legal_categories), downloaded)
                    else:
                        downloaded = backend._download_legal_categories()
                        self.assertEqual(sorted(backend.legal_categories), downloaded)

    def test_long_description(self):
        """Test long descriptions."""
        unlimited_length = 50000  # use this if the backend sets no limit
        for cls in Backend.all_backend_classes(needs={'scan'}):
            with self.subTest(cls):
                with self.temp_backend(cls, count=1) as backend:
                    track = backend[0]
                    max_length = backend._max_length.get('description') or unlimited_length
                    # a backend may encode keywords in description
                    max_descr_length = max_length - (len(backend._encode_description(track)) - len(track.description))
                    track.description = ('long description' * 4000)[:max_descr_length]
                    self.assertEqual(len(backend._encode_description(track)), max_length)
                    clone = backend.clone()[0]
                    self.assertEqual(track.description, clone.description)
                    if max_length < unlimited_length:
                        try_description = 'long description' * 4000
                        encoded = backend._encode_description(track)
                        decoded = backend._decode_description(track, encoded)
                        self.assertEqual(len(encoded), max_length)
                        self.assertTrue(try_description.startswith(decoded))

    def test_no_auth(self):
        """Some backends must fail if given no login data."""
        for cls in Backend.all_backend_classes(needs={'scan'}):
            if not cls.needs_config:
                continue
            for _ in ({'username': 'gpxitytest', 'password': ''}, {'username': ''}, {}, {'password': 'test'}):
                with self.assertRaises(Backend.BackendException) as context:
                    with self.temp_backend(cls, username=_):
                        pass
                self.assertEqual(str(context.exception), '{}: Needs authentication data'.format(cls.default_url))



# -*- coding: utf-8 -*-


# Copyright (c) 2018 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""Tests for gpxity.backends."""

import unittest
import pwd
import os
import io
import datetime
import time
import random
from pkgutil import get_data
import tempfile
from contextlib import contextmanager
from subprocess import Popen, PIPE
import logging
try:
    import MySQLdb
    import _mysql_exceptions
except ImportError:
    pass


import gpxpy
from gpxpy.gpx import GPXTrackPoint

from ... import Track
from ...auth import Authenticate
from .. import Mailer, WPTrackserver, Directory, GPSIES

# pylint: disable=attribute-defined-outside-init,protected-access

__all__ = ['BasicTest']


def disabled(*args) ->bool:
    """True if any of the backends is disabled.

    Returns:
        True or False

    """
    reason = list()
    for _ in args:
        if _.is_disabled():
            reason.append(_.__name__)
    return bool(reason), '{} {} disabled'.format(','.join(reason), 'is' if len(reason) == 1 else 'are')  # noqa


class BasicTest(unittest.TestCase):

    """define some helpers."""

    test_passwd = 'pwd'

    mysql_ip_address = None
    mysql_docker_name = 'gpxitytest_mysql'

    def setUp(self):  # noqa
        """define test specific Directory.prefix."""
        self.maxDiff = None  # pylint: disable=invalid-name
        Authenticate.path = os.path.join(os.path.dirname(__file__), 'test_auth_cfg')
        self.logger = logging.getLogger()
        self.logger.level = logging.DEBUG
        self.logger.debug('auth file now is %s', Authenticate.path)
        self.start_time = datetime.datetime.now()
        self.unicode_string1 = 'unicode szlig: ß'
        self.unicode_string2 = 'something japanese:の諸問題'
        Directory.prefix = 'gpxity.' + '.'.join(self.id().split('.')[-2:]) + '_'  # noqa
        path = tempfile.mkdtemp(prefix=Directory.prefix)

        if not os.path.exists(path):
            os.mkdir(path)
        Directory.prefix = path

        if not Mailer.is_disabled():
            self.start_mailserver()

    def tearDown(self):  # noqa
        """Check if there are still /tmp/gpxitytest.* directories."""
        if not Mailer.is_disabled():
            self.stop_mailserver()
        os.rmdir(Directory.prefix)
        timedelta = datetime.datetime.now() - self.start_time
        self.logger.debug('%s seconds ', timedelta.seconds)
        logging.shutdown()

    @contextmanager
    def subTest(self, name, **params):  # noqa pylint: disable=arguments-differ
        """With pytest, subTest does not do much. At least print the name."""
        if not isinstance(name, str):
            name = name.__name__
        _ = '{} subTest {}: {}'.format('-' * 10, self.id().split('.')[-1], name)
        _ += '-' * (80 - len(_))
        self.logger.debug(_)
        yield super(BasicTest, self).subTest(' ' + name, **params)

    @staticmethod
    def _get_gpx_from_test_file(name: str):
        """get data from a predefined gpx file.

        name is without .gpx

        Returns:=
            A GPX object

        """
        gpx_test_file = os.path.join(os.path.dirname(__file__), '{}.gpx'.format(name))
        if not os.path.exists(gpx_test_file):
            raise Exception('MMTTests needs a GPX file named {}.gpx for testing in {}'.format(
                name, os.getcwd()))
        return gpxpy.parse(io.StringIO(get_data(__package__, '{}.gpx'.format(name)).decode('utf-8')))

    @classmethod
    def create_test_track(
            cls, count: int = 1, idx: int = 0, category: str = None, public: bool = False,
            start_time=None, end_time=None):
        """create a :class:`~gpxity.Track`.

        It starts off with **test.gpx** and appends a
        last track point, it also changes the time stamp of the last point.
        This is done using **count** and **idx**: The last point is set such that
        looking at the tracks, they all go in a different direction clockwise, with an angle
        in degrees of :literal:`360 * idx / count`.

        Args:
            count: See above. Using 1 as default if not given.
            idx: See above. Using 0 as default if not given.
            category: The wanted value for the track.
                Default: if count == len(:attr:`Track.legal_categories <gpxity.Track.legal_categories>`),
                the default value will be legal_categories[idx].
                Otherwise a random value will be applied.
            public: should the tracks be public or private?
            start_time: If given, assign it to the first point and adjust all following times
            end_time: explicit time for the last point. If None: See above.

        Returns:
            (~gpxity.Track): A new track not bound to a backend

        """
        gpx = cls._get_gpx_from_test_file('test')
        if start_time is not None:
            _ = start_time - gpx.tracks[0].segments[0].points[0].time
            gpx.adjust_time(_)
        last_points = gpx.tracks[-1].segments[-1].points
        if end_time is None:
            end_time = last_points[-1].time + datetime.timedelta(hours=10, seconds=idx)
        new_point = GPXTrackPoint(
            latitude=last_points[-1].latitude, longitude=last_points[-1].longitude + 0.001, time=end_time)
        _ = gpxpy.geo.LocationDelta(distance=1000, angle=360 * idx / count)
        new_point.move(_)
        last_points.append(new_point)

        # now set all times such that they are in order with this track and do not overlap
        # with other test tracks
        _ = gpx.tracks[0].segments[0].points[0].time
        duration = new_point.time - _ + datetime.timedelta(seconds=10)
        for point in gpx.walk(only_points=True):
            point.time += duration * idx

        result = Track(gpx=gpx)
        result.title = 'Random GPX # {}'.format(idx)
        result.description = 'Description to {}'.format(gpx.name)
        if category:
            result.category = category
        elif count == len(Track.legal_categories):
            result.category = Track.legal_categories[idx]
        else:
            result.category = random.choice(Track.legal_categories)
        result.public = public
        return result

    @staticmethod
    def _random_datetime():
        """random datetime between now() - 10 days and now().

        Returns:
            A random datetime

        """
        end = datetime.datetime.now()
        start = end - datetime.timedelta(days=10)
        delta = end - start
        int_delta = (delta.days * 24 * 60 * 60) + delta.seconds
        random_second = random.randrange(int_delta)
        return start + datetime.timedelta(seconds=random_second)

    @staticmethod
    def _random_keywords(count=100):
        """A set of random keywords, but always the same.

        We do not want to generate too many tag ids for MMT.

        Returns:
            A set of random keywords

        """
        state = random.getstate()
        try:
            random.seed(1)
            basis = 'abcdefghijklmnopqrstuvwxyz'
            basis += basis.upper()
            basis += '/-_+.% $"|\\'
            result = set()
            while len(result) < count:
                candidate = ''.join(random.choice(basis) for x in range(4))
                if candidate[0] not in ' -' and candidate[-1] not in ' ':
                    result.add(candidate)
            return result
        finally:
            random.setstate(state)

    @classmethod
    def _random_points(cls, count=100):
        """Get some random points.

        Returns:
            A list with count points

        """
        result = list()
        start_time = cls._random_datetime()
        for _ in range(count):
            point = GPXTrackPoint(
                latitude=random.uniform(0.0, 90.0),
                longitude=random.uniform(0.0, 180.0), elevation=_,
                time=start_time + datetime.timedelta(seconds=10 * _))
            result.append(point)
        return result

    def assertSameTracks(self, backend1, backend2, msg=None, with_category=True):  # noqa pylint: disable=invalid-name
        """both backends must hold identical tracks."""
        self.maxDiff = None  # pylint: disable=invalid-name
        if backend1 != backend2:
            with_last_time = not (isinstance(backend1, GPSIES) or isinstance(backend2, GPSIES))
            keys1 = sorted(x.key(with_category, with_last_time) for x in backend1)
            keys2 = sorted(x.key(with_category, with_last_time) for x in backend2)
            self.assertEqual(keys1, keys2, msg)

    def assertEqualTracks(self, track1, track2, msg=None, xml: bool = False, with_category: bool = True):  # noqa pylint: disable=invalid-name
        """both tracks must be identical. We test more than necessary for better test coverage.

        Args:

            xml: if True, also compare to_xml()"""
        self.maxDiff = None

        # GPSIES: when uploading tracks. GPSIES sometimes assigns new times to all points,
        # starting at 2010-01-01 00:00. Until I find the reason, ignore point times for comparison.
        with_last_time = not (isinstance(track1.backend, GPSIES) or isinstance(track2.backend, GPSIES))

        self.assertEqual(track1.key(with_category, with_last_time), track2.key(with_category, with_last_time), msg)
        self.assertTrue(track1.points_equal(track2), msg)
        if xml:
            self.assertEqual(track1.gpx.to_xml(), track2.gpx.to_xml(), msg)

    def assertNotEqualTracks(self, track1, track2, msg=None, with_category: bool = True):  # noqa pylint: disable=invalid-name
        """both tracks must be different. We test more than necessary for better test coverage."""
        self.assertNotEqual(track1.key(with_category), track2.key(with_category), msg)
        self.assertFalse(track1.points_equal(track2), msg)
        self.assertNotEqual(track1.gpx.to_xml(), track2.gpx.to_xml(), msg)

    def assertTrackFileContains(self, track, string, msg=None):  # noqa pylint: disable=invalid-name
        """Assert that string is in the physical file. Works only for Directory backend."""
        with open(track.backend.gpx_path(track.id_in_backend), encoding='utf8') as trackfile:
            data = trackfile.read()
        self.assertIn(string, data, msg)

    def setup_backend(  # pylint: disable=too-many-arguments
            self, cls_, username: str = None, url: str = None, count: int = 0,
            cleanup: bool = True, clear_first: bool = True, category: str = None,
            public: bool = False):
        """set up an instance of a backend with count tracks.

        If count == len(:attr:`Track.legal_categories <gpxity.Track.legal_categories>`),
        the list of tracks will always be identical. For an example
        see :meth:`TestBackends.test_all_category <gpxity.backends.test.test_backends.TestBackends.test_all_category>`.

        Args:
            cls_ (Backend): the class of the backend to be created
            username: use this to for a specific accout name. Default is 'gpxitytest'.
                Special case WPTrackserver: pass the IP address of the mysql test server
            url: for the backend
            count: how many random tracks should be inserted?
            cleanup: If True, remove all tracks when done. Passed to the backend.
            clear_first: if True, first remove all existing tracks
            public: should the tracks be public or private?

        Returns:
            the prepared Backend

        """

        if username is None:
            username = 'gpxitytest'

        if cls_ is WPTrackserver:
            self.create_temp_mysqld()
            auth = {
                'Mysql': 'root@gpxitytest_db',
                'Password': self.test_passwd,
                'Url': self.mysql_ip_address,
                'Username': username}
            url = self.mysql_ip_address
        elif cls_ is Mailer:
            auth = {
                'Username': username,
                'port': 8025,
                'url': pwd.getpwuid(os.geteuid()).pw_name}
        else:
            auth = username
        result = cls_(url, auth=auth, cleanup=cleanup)
        if clear_first and'scan' in cls_.supported and 'write' in cls_.supported:
            result.remove_all()
        if count:
            # if count == 0, skip this. Needed for write-only backends like Mailer.
            while count > len(result):
                track = self.create_test_track(count, len(result), category=category, public=public)
                result.add(track)
            self.assertGreaterEqual(len(result), count)
            if clear_first:
                self.assertEqual(len(result), count)
        return result

    @staticmethod
    @contextmanager
    def lifetrackserver(directory, servername, port):
        """Start and ends a server for lifetrack testing."""
        exec_name = 'bin/gpxity_server'
        logfile = os.path.join(directory, 'gpxity_server.log')
        if not os.path.exists(exec_name):
            exec_name = 'gpxity_server'
        cmdline = '{} --loglevel debug --servername {} --port {} --directory {}'.format(
            exec_name, servername, port, directory)
        if not Mailer.is_disabled():
            cmdline += ' --mailto {} --smtp-port 8025'.format(pwd.getpwuid(os.geteuid()).pw_name)
        process = Popen(cmdline.split(), stderr=open(logfile, 'a'))
        try:
            time.sleep(1)  # give the server time to start
            yield
        finally:
            if os.path.exists(logfile):
                for _ in open(logfile):
                    logging.debug('SRV: %s', _.rstrip())
                os.remove(logfile)
            elif os.path.exists(directory):
                logging.debug('SRV: Directory exists but not gpxity_server.log')
            else:
                logging.debug('SRV: Directory %s does not exist', directory)
            process.kill()

    def start_mailserver(self):
        """Start an smptd server for mail testing."""
        self.mailserver_process = Popen(
            'aiosmtpd -u -n -d -l 127.0.0.1:8025'.split(),
            stdout=open('{}/smtpd_stdout'.format(Directory.prefix), 'w'),
            stderr=open('{}/smtpd_stderr'.format(Directory.prefix), 'w'))
        time.sleep(1)  # give the server time to start

    def stop_mailserver(self):
        """Stop the smtp server for mail testing."""
        self.mailserver_process.kill()
        for _ in ('out', 'err'):
            filename = '{}/smtpd_std{}'.format(Directory.prefix, _)
            if not os.path.exists(filename):
                logging.debug('MAIL: %s not found', filename)
                continue
            if _ == 'err':
                for fileline in open(filename):
                    if not fileline.startswith('INFO:'):
                        logging.debug('MAIL:%s: %s', _, fileline.rstrip())
            os.remove(filename)

    @contextmanager
    def temp_backend(self, cls_, url=None, count=0,  # pylint: disable=too-many-arguments
                     cleanup=True, clear_first=True, category=None,
                     public: bool = False, username=None):
        """Just like setup_backend but usable as a context manager. which will call destroy() when done."""
        tmp_backend = self.setup_backend(cls_, username, url, count, cleanup, clear_first, category, public)
        try:
            yield tmp_backend
        finally:
            tmp_backend.destroy()

    @classmethod
    def create_db_for_wptrackserver(cls):
        """Create the mysql database for the WPTrackserver tests."""
        while True:
            try:
                server = MySQLdb.connect(
                    host=cls.mysql_ip_address, user='root', passwd=cls.test_passwd, connect_timeout=2,
                    autocommit=True, charset='utf8',
                    sql_mode='ONLY_FULL_GROUP_BY,STRICT_TRANS_TABLES,NO_ENGINE_SUBSTITUTION')
                break
            except _mysql_exceptions.OperationalError:
                # wait until the docker instance is ready
                time.sleep(1)
        logging.debug('%s: connected to %s', datetime.datetime.now(), cls.mysql_ip_address)
        cursor = server.cursor()
        cursor.execute('create database gpxitytest_db')
        cursor.execute('use gpxitytest_db')
        cursor.execute("""
            CREATE TABLE wp_ts_locations (
            id int(11) NOT NULL AUTO_INCREMENT,
            trip_id int(11) NOT NULL,
            latitude double NOT NULL,
            longitude double NOT NULL,
            altitude double NOT NULL,
            speed double NOT NULL default 0.0,
            heading double NOT NULL default 0.0,
            updated timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
            created timestamp NOT NULL default '0000-00-00 00:00:00',
            occurred timestamp not null default '0000-00-00 00:00:00',
            comment varchar(255) NOT NULL,
            hidden tinyint(1) NOT NULL default 0,
            PRIMARY KEY (id),
            KEY occurred (occurred),
            KEY trip_id (trip_id)
            ) ENGINE=InnoDB
        """)
        cursor.execute("""
            CREATE TABLE wp_ts_tracks (
            id int(11) NOT NULL AUTO_INCREMENT,
            user_id int(11) NOT NULL,
            name varchar({title_length}) NOT NULL,
            updated timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
            created timestamp NOT NULL default '0000-00-00 00:00:00',
            source varchar(255) NOT NULL default '',
            comment varchar({descr_length}) NOT NULL default '',
            distance int(11) NOT NULL,
            PRIMARY KEY (id),
            KEY user_id (user_id)
         )
        """.format(
            descr_length=WPTrackserver._max_length['description'],
            title_length=WPTrackserver._max_length['title']))
        cursor.execute("""
            CREATE TABLE wp_users (
            ID bigint(20) unsigned NOT NULL AUTO_INCREMENT,
            user_login varchar(60) NOT NULL DEFAULT '',
            user_pass varchar(255) NOT NULL DEFAULT '',
            user_nicename varchar(50) NOT NULL DEFAULT '',
            user_email varchar(100) NOT NULL DEFAULT '',
            user_url varchar(100) NOT NULL DEFAULT '',
            user_registered datetime NOT NULL DEFAULT '1970-01-01',
            user_activation_key varchar(255) NOT NULL DEFAULT '',
            user_status int(11) NOT NULL DEFAULT '0',
            display_name varchar(250) NOT NULL DEFAULT '',
            PRIMARY KEY (ID),
            KEY user_login_key (user_login),
            KEY user_nicename (user_nicename),
            KEY user_email (user_email)
         )
        """)
        cursor.execute("""
            insert into wp_users (user_login) values(%s)""", ['gpxitytest'])
        cursor.execute("select id,user_login from wp_users")
        logging.debug(cursor.fetchall())
        server.commit()
        server.close()

    @classmethod
    def find_mysql_docker(cls) ->bool:
        """Find an already running docker.

        Returns:
            the IP address

        """
        if cls.mysql_ip_address is not None:
            return True
        cls.mysql_ip_address = Popen([
            'docker', 'inspect', '--format', '{{range .NetworkSettings.Networks}}{{.IPAddress}}{{end}}',
            cls.mysql_docker_name], stdout=PIPE).communicate()[0].decode().strip()
        if cls.mysql_ip_address == '':
            cls.mysql_ip_address = None
        logging.debug('%s at ip %s', cls.mysql_docker_name, cls.mysql_ip_address)
        return bool(cls.mysql_ip_address)

    @classmethod
    def create_temp_mysqld(cls):
        """Create a temporary mysql server and initialize it with test data for WPTrackserver."""
        if cls.find_mysql_docker():
            return
        cmd = [
            'docker', 'run', '--name', cls.mysql_docker_name, '--detach',
            '--env', 'MYSQL_ROOT_PASSWORD={}'.format(cls.test_passwd), 'mysql']
        with Popen(cmd, stdout=PIPE, stderr=PIPE) as _:
            std_err = _.stderr.read().strip().decode('utf8')
            if std_err:
                logging.error(std_err)
                if 'on network bridge: failed to add the host' in std_err:
                    logging.error('did you reboot after kernel upgrade?')
                raise Exception('Cannot run docker: {}'.format(std_err))
        if not cls.find_mysql_docker():
            raise Exception('Unknown problem while creating mysql docker')
        cls.create_db_for_wptrackserver()
